"""
    mosaic_properties
    =================

    Properties of an asset.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['MosaicProperties']


DTO2Type = typing.Sequence[dict]
FLAGS_ID = 0
DIVISIBILITY_ID = 1
DURATION_ID = 2
PROPERTY_SIZE = 9
PROPERTIES = {
    FLAGS_ID: 'flags',
    DIVISIBILITY_ID: 'divisibility',
    DURATION_ID: 'duration',
}


def to_flags(
    supply_mutable: bool,
    transferable: bool,
    levy_mutable: bool
) -> int:
    """Convert 3 binary values to sequential bit flags"""

    flags: int = 0
    if supply_mutable:
        flags |= 1
    if transferable:
        flags |= 2
    if levy_mutable:
        flags |= 4

    return flags


def property_to_catbuffer(id: int, value: int) -> bytes:
    """Convert property with ID to catbuffer."""
    return util.u8_to_catbuffer(id) + util.u64_to_catbuffer(value)


def property_from_catbuffer(catbuffer: bytes) -> typing.Tuple[int, int]:
    """Convert catbuffer to property with ID."""

    id = util.u8_from_catbuffer(catbuffer[:1])
    value = util.u64_from_catbuffer(catbuffer[1:])
    return (id, value)


@util.inherit_doc
@util.dataclass(frozen=True)
class MosaicProperties(util.DTO):
    """
    Properties of an asset.

    Note: The `MosaicDefinitionTransaction` uses a different DTO
    format for `MosaicProperties` than specified here.

    :param flags: Flags for the properties of the mosaic.
    :param divisibility: Decimal places mosaic can be divided into [0-6].
    :param duration: Number of blocks the mosaic will be available.

    DTO Format:
        .. code-block:: yaml

            MosaicPropertiesDTO: UInt64DTO[]
    """

    flags: int
    divisibility: int
    duration: int

    def __init__(
        self,
        flags: int,
        divisibility: int = 0,
        duration: int = 0,
    ):
        if flags < 0 or flags > 7:
            raise ValueError('Invalid flags, not in range [0-7].')
        if divisibility < 0 or divisibility > 6:
            raise ValueError('Invalid divisibility, not in range [0-6].')
        self._set('flags', flags)
        self._set('divisibility', divisibility)
        self._set('duration', duration)

    @property
    def supply_mutable(self) -> bool:
        """Mosaic allows a supply change later on. Default false."""
        return (self.flags & 1) == 1

    @property
    def transferable(self) -> bool:
        """Allow transfer of funds from non-creator accounts. Default true."""
        return (self.flags & 2) == 2

    @property
    def levy_mutable(self) -> bool:
        """Get if levy is mutable. Default false."""
        return (self.flags & 4) == 4

    @classmethod
    def create(cls, **kwds):
        """
        Create mosaic properties with default parameters.

        :param supply_mutable: Mosaic allows supply change later.
        :param transferable: Allow transfer of funds from accounts other than creator.
        :param levy_mutable: If levy is mutable.
        :param divisibility: Decimal place mosaic can be divided into.
        :param duration: Number of blocks the mosaic will be available.
        """
        supply_mutable = typing.cast(bool, kwds.get('supply_mutable', False))
        transferable = typing.cast(bool, kwds.get('transferable', True))
        levy_mutable = typing.cast(bool, kwds.get('levy_mutable', False))
        divisibility = kwds.get('divisibility', 0)
        duration = kwds.get('duration', 0)
        flags = to_flags(supply_mutable, transferable, levy_mutable)
        return cls(flags, divisibility, duration)

    @classmethod
    def validate_dto(cls, data: typing.Sequence[dict]) -> bool:
        """Validate the data-transfer object."""

        return (
            len(data) in {2, 3}
            and all(len(i) == 2 for i in data)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> typing.Sequence[dict]:
        # For indefinite mosaics, the duration is optional (default 0).
        return [
            {'id': 0, 'value': util.u64_to_dto(self.flags)},
            {'id': 1, 'value': util.u64_to_dto(self.divisibility)},
            {'id': 2, 'value': util.u64_to_dto(self.duration)},
        ]

    @classmethod
    def create_from_dto(
        cls,
        data: typing.Sequence[dict],
        network_type: OptionalNetworkType = None,
    ) -> MosaicProperties:
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        # For indefinite mosaics, the duration is optional (default 0).
        duration = 0

        for prop in data[0:3]:
            prop_id = util.u8_from_dto(prop["id"])

            if (prop_id == 0):
                flags = util.u64_from_dto(prop["value"])
            elif (prop_id == 1):
                divisibility = util.u64_from_dto(prop["value"])
            elif (prop_id == 2):
                duration = util.u64_from_dto(prop["value"])
            else:
                raise ValueError('Invalid data-transfer object.')

        return cls(flags, divisibility, duration)

    # TESTING
    # Private, internal helper for testing only.

    def _permute_(self, cb) -> MosaicProperties:
        """Permute data inside self."""

        flags = 0x7 - self.flags
        divisibility = 6 - self.divisibility
        duration = cb(self.duration)
        return MosaicProperties(flags, divisibility, duration)


@util.inherit_doc
@util.dataclass(frozen=True)
class MosaicDefinitionProperties(util.Model):
    """Internal class to provide transaction support for mosaic properties."""

    model: MosaicProperties

    def catbuffer_size(self) -> int:
        """Get the mosaic properties size as catbuffer."""

        # Get the number of optional properties.
        count = 0
        if self.model.duration != 0:
            count += 1

        # uint8_t count
        # uint8_t flags
        # uint8_t divisibility
        # typedef MosaicProperty { id: uint8_t, value: uint64_t }
        # MosaicProperty[count] properties
        count_size = util.U8_BYTES
        flags_size = util.U8_BYTES
        divisibility_size = util.U8_BYTES
        property_size = util.U64_BYTES + util.U8_BYTES
        return count_size + flags_size + divisibility_size + count * property_size

    @classmethod
    def validate_dto(cls, data: DTO2Type) -> bool:
        """Validate the data-transfer object."""

        required_keys = {'id', 'value'}
        return (
            len(data) >= 2
            and all((
                cls.validate_dto_required(entry, required_keys)
                and cls.validate_dto_all(entry, required_keys)
            ) for entry in data)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ):
        # A newer version of DTO, which is used in MosaicDefinitionTransactions.
        # We need to keep the two versions separate.
        data = [
            {'id': FLAGS_ID, 'value': util.u64_to_dto(self.model.flags)},
            {'id': DIVISIBILITY_ID, 'value': util.u64_to_dto(self.model.divisibility)},
        ]
        if self.model.duration != 0:
            data.append({
                'id': DURATION_ID,
                'value': util.u64_to_dto(self.model.duration)
            })

        return data

    @classmethod
    def create_from_dto(
        cls,
        data: DTO2Type,
        network_type: OptionalNetworkType = None,
    ):
        # There's a data inconsistnecy in reply from node
        # /mosaic routes contains 'id' field
        # MosaciDefinition transactions replies with 'key' field
        for item in data:
            if (('key' in item) and ('id' not in item)):
                item['id'] = item.pop('key')

        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        # A newer version of DTO, which is used in MosaicDefinitionTransactions.
        # We need to keep the two versions separate.
        kwds = {}
        for item in data:
            kwds[PROPERTIES[item['id']]] = util.u64_from_dto(item['value'])
        return cls(MosaicProperties(**kwds))

    def to_catbuffer(
        self,
        network_type: OptionalNetworkType = None,
        fee_strategy: typing.Optional[util.FeeCalculationStrategy] = util.FeeCalculationStrategy.MEDIUM,
    ) -> bytes:
        # Serialize the required properties.
        flags = util.u8_to_catbuffer(self.model.flags)
        divisibility = util.u8_to_catbuffer(self.model.divisibility)

        # Serialize the optional properties.
        counter = 0
        optional = b''
        if self.model.duration != 0:
            counter += 1
            optional += property_to_catbuffer(DURATION_ID, self.model.duration)

        count = util.u8_to_catbuffer(counter)
        return count + flags + divisibility + optional

    @classmethod
    def create_from_catbuffer_pair(
        cls,
        data: bytes,
        network_type: OptionalNetworkType = None,
    ):
        # Read the array count, property flags and divisibility.
        count = util.u8_from_catbuffer(data[:1])
        flags = util.u8_from_catbuffer(data[1:2])
        divisibility = util.u8_from_catbuffer(data[2:3])

        # Ensure the buffer is long enough for the data, and iteratively
        # read the remaining properties.
        step = PROPERTY_SIZE
        property_size = step * count
        size = 3 + property_size
        kwds = {}
        for i in range(0, property_size, step):
            start = 3 + i
            stop = start + step
            id, value = property_from_catbuffer(data[start:stop])
            kwds[PROPERTIES[id]] = value

        # Instantiate our class and return pair.
        inst = cls(MosaicProperties(flags, divisibility, **kwds))
        remaining = data[size:]
        return inst, remaining



"""
    namespace_info
    ==============

    Data and metadata describing a namespace.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

from .alias import Alias
from .alias_type import AliasType
from .namespace_id import NamespaceId, NamespaceIdList
from .namespace_type import NamespaceType
from ..account.address import Address
from ..account.public_account import PublicAccount
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['NamespaceInfo']


@util.inherit_doc
@util.dataclass(frozen=True, alias=Alias(AliasType.NONE, None))
class NamespaceInfo(util.DTO):
    """
    Information describing a namespace.

    :param active: Namespace is active.
    :param index: Namespace index.
    :param meta_id: Metadata ID.
    :param type: Namespace type.
    :param depth: Level of namespace.
    :param levels: Namespace ID levels.
    :param parent_id: Namespace parent ID.
    :param owner: Account that owns namespace.
    :param start_height: Block height at which ownership begins.
    :param end_height: Block height at which ownership ends.
    :param alias: Alias linked to a namespace.

    DTO Format:
        .. code-block:: yaml

            NamespaceMetaDTO:
                # Hex(Id) (24-bytes)
                id: string
                active: boolean
                index: integer

            NamespaceDTO:
                # Hex(PublicKey) (64-bytes)
                owner: string
                # Hex(Address) (50-bytes)
                ownerAddress: string
                startHeight: UInt64DTO
                endHeight: UInt64DTO
                depth: integer
                level0: UInt64DTO
                level1?: UInt64DTO
                level2?: UInt64DTO
                type: integer
                # Required, but made optional for backward compatibility.
                alias?: AliasDTO
                parentId: UInt64DTO

            NamespaceInfoDTO:
                meta: NamespaceMetaDTO
                namespace: NamespaceDTO
    """

    active: bool
    index: int
    meta_id: str
    type: NamespaceType
    depth: int
    levels: NamespaceIdList
    parent_id: NamespaceId
    owner: PublicAccount
    start_height: int
    end_height: int
    alias: Alias

    @property
    def id(self) -> NamespaceId:
        """Get the namespace ID."""
        return self.levels[-1]

    def is_root(self) -> bool:
        """Get if namespace is root namespace."""

        return self.type == NamespaceType.ROOT_NAMESPACE

    def is_subnamespace(self) -> bool:
        """Get if namespace is subnamespace."""

        return self.type == NamespaceType.SUB_NAMESPACE

    def has_alias(self) -> bool:
        """Get if namespace has alias."""

        return self.alias.type != AliasType.NONE

    def parent_namespace_id(self) -> NamespaceId:
        """Get parent namespace ID."""

        if self.is_root():
            raise ValueError("Unable to get parent namespace ID of root namespace.")
        return self.parent_id

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        meta = {
            'active': self.active,
            'index': self.index,
            'id': self.meta_id,
        }
        namespace = {
            'type': self.type.to_dto(network_type),
            'depth': self.depth,
            'parentId': util.u64_to_dto(int(self.parent_id)),
            'owner': self.owner.public_key,
            'ownerAddress': util.hexlify(self.owner.address.encoded),
            'startHeight': util.u64_to_dto(self.start_height),
            'endHeight': util.u64_to_dto(self.end_height),
        }

        # levels => ('level0', 'level1', ...)
        for i in range(self.depth):
            namespace[f'level{i}'] = util.u64_to_dto(int(self.levels[i]))

        # Optional alias.
        if self.alias.type != AliasType.NONE:
            namespace['alias'] = self.alias.to_dto(network_type)

        return {
            'meta': meta,
            'namespace': namespace,
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        meta = data['meta']
        namespace = data['namespace']
        address = Address.create_from_encoded(namespace['ownerAddress'])
        network_type = address.network_type
        depth = namespace['depth']

        # Load the levels.
        levels = []
        for i in range(depth):
            levels.append(NamespaceId(util.u64_from_dto(namespace[f'level{i}'])))

        # Load the alias.
        try:
            alias = Alias.create_from_dto(namespace['alias'], network_type)
        except KeyError:
            alias = Alias(AliasType.NONE, None)

        return cls(
            active=meta['active'],
            index=meta['index'],
            meta_id=meta['id'],
            type=NamespaceType.create_from_dto(namespace['type'], network_type),
            depth=depth,
            levels=levels,
            parent_id=NamespaceId(util.u64_from_dto(namespace['parentId'])),
            owner=PublicAccount(address, namespace['owner']),
            start_height=util.u64_from_dto(namespace['startHeight']),
            end_height=util.u64_from_dto(namespace['endHeight']),
            alias=alias,
        )



"""
    property_type
    =============

    Types of properties for an account.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import enum

from ... import util

__all__ = ['PropertyType']


@util.inherit_doc
class PropertyType(util.U8Mixin, util.EnumMixin, enum.IntEnum):
    """Identifier for an account property type."""

    ALLOW_ADDRESS       = 0x01  # noqa: E221
    ALLOW_MOSAIC        = 0x02  # noqa: E221
    ALLOW_TRANSACTION   = 0x04  # noqa: E221
    SENTINEL            = 0x05  # noqa: E221
    BLOCK_ADDRESS       = 0x81  # noqa: E221
    BLOCK_MOSAIC        = 0x82  # noqa: E221
    BLOCK_TRANSACTION   = 0x84  # noqa: E221

    def description(self) -> str:
        return DESCRIPTION[self]


DESCRIPTION = {
    PropertyType.ALLOW_ADDRESS: "The property type is an address.",
    PropertyType.ALLOW_MOSAIC: "The property type is a mosaic id.",
    PropertyType.ALLOW_TRANSACTION: "The property type is a transaction type.",
    PropertyType.SENTINEL: "Property type sentinel.",
    PropertyType.BLOCK_ADDRESS: "The property type is a blocking address operation.",
    PropertyType.BLOCK_MOSAIC: "The property type is a blocking mosaic id operation.",
    PropertyType.BLOCK_TRANSACTION: "The property type is a blocking transaction type operation.",  # noqa: E501
}



"""
    multsig_cosignatory_modification
    ================================

    Multisig cosignatory modification.

    Enables the addition or deletion of a cosignatory from a multisig account.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .multisig_cosignatory_modification_type import MultisigCosignatoryModificationType
from ..account.public_account import PublicAccount
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['MultisigCosignatoryModification']

ModificationType = MultisigCosignatoryModificationType


@util.inherit_doc
@util.dataclass(frozen=True)
class MultisigCosignatoryModification(util.Model):
    """
    Multisig cosignatory modification.

    :param type: Multi-signature modification type.
    :param cosignatory_public_account: Cosignatory public account.
    """

    type: ModificationType
    cosignatory_public_account: PublicAccount
    CATBUFFER_SIZE: typing.ClassVar[int] = 33 * util.U8_BYTES

    def __init__(
        self,
        cosignatory_public_account: PublicAccount,
        type: ModificationType
    ) -> None:
        self._set('cosignatory_public_account', cosignatory_public_account)
        self._set('type', type)

    @classmethod
    def create(
        cls,
        cosignatory_public_account: PublicAccount,
        type: ModificationType
    ):
        return cls(
            cosignatory_public_account,
            type
        )

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys = {'cosignatoryPublicKey', 'type'}
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, required_keys)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'cosignatoryPublicKey': self.cosignatory_public_account.public_key,
            'type': self.type.to_dto(network_type),
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        public_key = data['cosignatoryPublicKey']
        type = ModificationType.create_from_dto(data['type'], network_type)
        public_account = PublicAccount.create_from_public_key(public_key, network_type)
        return cls(public_account, type)

    def to_catbuffer(
        self,
        network_type: OptionalNetworkType = None,
        fee_strategy: util.FeeCalculationStrategy = util.FeeCalculationStrategy.MEDIUM,
    ) -> bytes:
        type = self.type.to_catbuffer(network_type)
        cosignatory = util.unhexlify(self.cosignatory_public_account.public_key)
        return type + cosignatory

    @classmethod
    def create_from_catbuffer(
        cls,
        data: bytes,
        network_type: OptionalNetworkType = None,
    ):
        type, data = ModificationType.create_from_catbuffer_pair(data, network_type)
        cosignatory = PublicAccount.create_from_public_key(data[:32], network_type)
        return cls(cosignatory, type)



from ..models.transaction.transaction_status_error import TransactionStatusError
from ..models.transaction.deadline import Deadline
from ..models.account.address import Address

__all__ = ['TransactionError']


class TransactionError(Exception):

    hash: str
    status: str
    deadline: Deadline
    channel_name: str
    address: Address

    def __init__(self, tx: TransactionStatusError):
        self.hash = tx.hash
        self.status = tx.status
        self.deadline = tx.deadline
        self.channel_name = tx.channel_name
        self.address = tx.address

    def __str__(self):
        return self.status



"""
    default
    =======

    Synchronous and asynchronous NIS client using the default backend.

    The core HTTP client shares a global session, to share a connection
    pool to speed up requests.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import aiohttp
import requests
import websockets
import typing

from . import abc
from . import client
from .. import util
from ..models.blockchain.network_type import NetworkType

__all__ = [
    # Synchronous
    'HTTP',
    'AccountHTTP',
    'BlockchainHTTP',
    'ContractHTTP',
    'MetadataHTTP',
    'ConfigHTTP',
    'NodeHTTP',
    'MosaicHTTP',
    'NamespaceHTTP',
    'NetworkHTTP',
    'TransactionHTTP',

    # Asynchronous
    'AsyncHTTP',
    'AsyncAccountHTTP',
    'AsyncBlockchainHTTP',
    'AsyncContractHTTP',
    'AsyncMetadataHTTP',
    'AsyncConfigHTTP',
    'AsyncNodeHTTP',
    'AsyncMosaicHTTP',
    'AsyncNamespaceHTTP',
    'AsyncNetworkHTTP',
    'AsyncTransactionHTTP',

    # Websockets
    'Listener',

    # Exceptions
    'HTTPError',
    'AsyncHTTPError',
]

# EXCEPTIONS

HTTPError = requests.HTTPError
AsyncHTTPError = aiohttp.ClientResponseError

# SYNCHRONOUS


@util.inherit_doc
class HTTPBase(abc.HTTPBase):
    """Abstract base class for synchronous HTTP clients."""

    def __init__(self, endpoint: str, network_type: typing.Optional[NetworkType] = None) -> None:
        self._endpoint = endpoint
        self._index = 0
        self._network_type = network_type

    def __enter__(self) -> HTTPBase:
        self._client = client.Client(requests.Session(), self._endpoint)
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        self.close()

    @property
    def root(self) -> HTTP:
        return HTTP.create_from_http(self)


@util.inherit_doc
class HTTP(HTTPBase, abc.HTTP):
    """Main client for the synchronous NIS API."""

    @property
    def account(self) -> AccountHTTP:
        return AccountHTTP.create_from_http(self)

    @property
    def blockchain(self) -> BlockchainHTTP:
        return BlockchainHTTP.create_from_http(self)

    @property
    def contract(self) -> ContractHTTP:
        return ContractHTTP.create_from_http(self)

    @property
    def metadata(self) -> MetadataHTTP:
        return MetadataHTTP.create_from_http(self)

    @property
    def config(self) -> ConfigHTTP:
        return ConfigHTTP.create_from_http(self)

    @property
    def node(self) -> NodeHTTP:
        return NodeHTTP.create_from_http(self)

    @property
    def mosaic(self) -> MosaicHTTP:
        return MosaicHTTP.create_from_http(self)

    @property
    def namespace(self) -> NamespaceHTTP:
        return NamespaceHTTP.create_from_http(self)

    @property
    def network(self) -> NetworkHTTP:
        return NetworkHTTP.create_from_http(self)

    @property
    def transaction(self) -> TransactionHTTP:
        return TransactionHTTP.create_from_http(self)


@util.inherit_doc
class AccountHTTP(HTTPBase, abc.AccountHTTP):
    """Account client for the synchronous NIS API."""


@util.inherit_doc
class BlockchainHTTP(HTTPBase, abc.BlockchainHTTP):
    """Blockchain client for the synchronous NIS API."""


@util.inherit_doc
class ContractHTTP(HTTPBase, abc.ContractHTTP):
    """Contract client for the synchronous NIS API."""


@util.inherit_doc
class MetadataHTTP(HTTPBase, abc.MetadataHTTP):
    """Metadata client for the synchronous NIS API."""


@util.inherit_doc
class ConfigHTTP(HTTPBase, abc.ConfigHTTP):
    """Config client for the synchronous NIS API."""


@util.inherit_doc
class NodeHTTP(HTTPBase, abc.NodeHTTP):
    """Node client for the synchronous NIS API."""


@util.inherit_doc
class MosaicHTTP(HTTPBase, abc.MosaicHTTP):
    """Mosaic client for the synchronous NIS API."""


@util.inherit_doc
class NamespaceHTTP(HTTPBase, abc.NamespaceHTTP):
    """Namespace client for the synchronous NIS API."""


@util.inherit_doc
class NetworkHTTP(HTTPBase, abc.NetworkHTTP):
    """Network client for the synchronous NIS API."""


@util.inherit_doc
class TransactionHTTP(HTTPBase, abc.TransactionHTTP):
    """Transaction client for the synchronous NIS API."""


# ASYNCHRONOUS


@util.inherit_doc
class AsyncHTTPBase(abc.AsyncHTTPBase):
    """Abstract base class for synchronous HTTP clients."""

    def __init__(
        self,
        endpoint: str,
        loop: util.OptionalLoopType = None,
        network_type: typing.Optional[NetworkType] = None,
    ) -> None:
        self._endpoint = endpoint
        self._index = 1
        self._loop = loop
        self._session = aiohttp.ClientSession(loop=loop)
        self._network_type = network_type

    async def __aenter__(self) -> AsyncHTTPBase:
        self._client = client.AsyncClient(self._session, self._endpoint)
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self.close()

    @property
    def root(self) -> AsyncHTTP:
        return AsyncHTTP.create_from_http(self)


@util.inherit_doc
class AsyncHTTP(AsyncHTTPBase, abc.HTTP):
    """Main client for the synchronous NIS API."""

    @property
    def account(self) -> AsyncAccountHTTP:
        return AsyncAccountHTTP.create_from_http(self)

    @property
    def blockchain(self) -> AsyncBlockchainHTTP:
        return AsyncBlockchainHTTP.create_from_http(self)

    @property
    def contract(self) -> AsyncContractHTTP:
        return AsyncContractHTTP.create_from_http(self)

    @property
    def metadata(self) -> AsyncMetadataHTTP:
        return AsyncMetadataHTTP.create_from_http(self)

    @property
    def config(self) -> AsyncConfigHTTP:
        return AsyncConfigHTTP.create_from_http(self)

    @property
    def node(self) -> AsyncNodeHTTP:
        return AsyncNodeHTTP.create_from_http(self)

    @property
    def mosaic(self) -> AsyncMosaicHTTP:
        return AsyncMosaicHTTP.create_from_http(self)

    @property
    def namespace(self) -> AsyncNamespaceHTTP:
        return AsyncNamespaceHTTP.create_from_http(self)

    @property
    def network(self) -> AsyncNetworkHTTP:
        return AsyncNetworkHTTP.create_from_http(self)

    @property
    def transaction(self) -> AsyncTransactionHTTP:
        return AsyncTransactionHTTP.create_from_http(self)


@util.inherit_doc
class AsyncAccountHTTP(AsyncHTTPBase, abc.AccountHTTP):
    """Account client for the asynchronous NIS API."""


@util.inherit_doc
class AsyncBlockchainHTTP(AsyncHTTPBase, abc.BlockchainHTTP):
    """Blockchain client for the asynchronous NIS API."""


@util.inherit_doc
class AsyncContractHTTP(AsyncHTTPBase, abc.ContractHTTP):
    """Contract client for the asynchronous NIS API."""


@util.inherit_doc
class AsyncMetadataHTTP(AsyncHTTPBase, abc.MetadataHTTP):
    """Metadata client for the asynchronous NIS API."""


@util.inherit_doc
class AsyncConfigHTTP(AsyncHTTPBase, abc.ConfigHTTP):
    """Config client for the asynchronous NIS API."""


@util.inherit_doc
class AsyncNodeHTTP(AsyncHTTPBase, abc.NodeHTTP):
    """Node client for the asynchronous NIS API."""


@util.inherit_doc
class AsyncMosaicHTTP(AsyncHTTPBase, abc.MosaicHTTP):
    """Mosaic client for the asynchronous NIS API."""


@util.inherit_doc
class AsyncNamespaceHTTP(AsyncHTTPBase, abc.NamespaceHTTP):
    """Namespace client for the asynchronous NIS API."""


@util.inherit_doc
class AsyncNetworkHTTP(AsyncHTTPBase, abc.NetworkHTTP):
    """Network client for the asynchronous NIS API."""


@util.inherit_doc
class AsyncTransactionHTTP(AsyncHTTPBase, abc.TransactionHTTP):
    """Transaction client for the asynchronous NIS API."""


# WEBSOCKETS


@util.inherit_doc
class Listener(abc.Listener):
    """Asynchronous websockets-based listener."""

    def __init__(
        self,
        endpoint: str,
        loop: util.OptionalLoopType = None,
        network_type: typing.Optional[NetworkType] = None,
    ) -> None:
        url = client.parse_ws_url(endpoint)
        self._loop = loop
        self._conn = websockets.connect(url.url, loop=loop)
        self._network_type = network_type

    async def __aenter__(self) -> Listener:
        self._session = await self._conn.__aenter__()
        self._client = client.WebsocketClient(self._session)
        self._iter_ = self._client.__aiter__()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self._conn.__aexit__(None, None, None)



"""
    address
    =======

    A public identifier for an account and network type.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from ..blockchain.network_type import NetworkType
from ... import util

__all__ = ['Address']

T = typing.TypeVar('T')
U = typing.Sequence[T]


def chunks(collection: U, n: int) -> typing.Generator[U, None, None]:
    """Generate n-sized chunks from collection."""

    for i in range(0, len(collection), n):
        yield collection[i:i + n]


def calculate_checksum(address: bytes) -> bytes:
    """Calculate the checksum for the address."""

    checksum = util.hashlib.sha3_256(address[:21]).digest()[:4]
    return typing.cast(bytes, checksum)


def public_key_to_address(
    public_key: bytes,
    network_type: NetworkType
) -> bytes:
    """
    Convert public key to address.

    :param public_key: Raw bytes of public key.
    :param network_type: Network type.
    :return: Address bytes.
    """

    # step 1: keccak256 hash of the public key
    public_key_hash: bytes = util.hashlib.sha3_256(public_key).digest()

    # step 2: ripemd160 hash of (1)
    ripemd_hash: bytes = util.hashlib.ripemd160(public_key_hash).digest()

    # step 3: add network identifier byte in front of (2)
    address = bytearray(25)
    address[0] = int(network_type)
    address[1:21] = ripemd_hash

    # step 4: concatenate (3) and the checksum of (3)
    address[21:] = calculate_checksum(address)

    return address


def is_valid_address(address: bytes) -> bool:
    """Check if address is valid."""

    return typing.cast(bool, calculate_checksum(address) == address[21:])


@util.inherit_doc
@util.dataclass(frozen=True)
class Address(util.Object):
    """
    Account address for network type.

    An address is derived from the public key and network type, and
    uniquely identifies a NEM account.

    :param address: Base32-encoded, human-readable address.
    """

    address: str
    network_type: NetworkType

    def __init__(self, address: str) -> None:
        plain = address.strip().upper().replace('-', '')
        if len(plain) != 40:
            raise ValueError(f"{address} is not a valid raw address")
        network_type = NetworkType.create_from_raw_address(plain)
        self._set('address', plain)
        self._set('network_type', network_type)

    @property
    def encoded(self) -> bytes:
        """Get encoded address."""
        return util.b32decode(self.address)

    @property
    def hex(self) -> str:
        """Get hex address."""
        return util.encode_hex(self.encoded).upper()

    @classmethod
    def create_from_raw_address(cls, address: str) -> Address:
        """
        Create Address from human-readable, raw address.

        :param address: Base32-encoded, human-readable address.
        :return: Address object.
        """
        return cls(address)

    @classmethod
    def create_from_encoded(cls, address: typing.AnyStr) -> Address:
        """
        Create Address from encoded address.

        :param address: Base32-decoded address bytes or hex-encoded bytes.
        :return: Address object.
        """

        address = util.decode_hex(address, with_prefix=True)
        if len(address) != 25:
            raise ValueError(f"{address!r} is not a valid encoded address")
        raw_address = util.b32encode(address)
        return cls.create_from_raw_address(raw_address)

    @classmethod
    def create_from_public_key(
        cls,
        public_key: typing.AnyStr,
        network_type: NetworkType
    ):
        """
        Create Address from the public key and network type.

        :param public_key: Hex-encoded public key (with or without '0x' prefix).
        :param network_type: Network type for address.
        :return: Address object.
        """

        key = util.decode_hex(public_key, with_prefix=True)
        if len(key) != 32:
            raise ValueError(f"{public_key!r} is not a valid public key")
        address: bytes = public_key_to_address(key, network_type)

        return cls.create_from_encoded(address)

    def plain(self) -> str:
        """Get plain representation of address as base32-decoded string."""
        return self.address

    def pretty(self) -> str:
        """Get pretty representation of address as base32-decoded string."""
        return u'-'.join(chunks(self.address, 6))

    def is_valid(self) -> bool:
        """Check if address is valid."""
        return is_valid_address(self.encoded)



"""
    receipt_base
    ===========

    Abstract base class for transaction objects.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .format import DTOFormat
from .receipt_type import ReceiptType
from .receipt_version import ReceiptVersion
from ..blockchain.network_type import OptionalNetworkType
from ... import util


ReceiptBaseType = typing.TypeVar('ReceiptBaseType', bound='ReceiptBase')
TypeMap = typing.Mapping[ReceiptType, typing.Type[ReceiptBaseType]]


@util.inherit_doc
@util.dataclass(frozen=True)
class ReceiptBase(util.Model):
    """
    Abstract, shared transaction base class.

    :param type: Transaction type.
    :param network_type: Network type.
    :param version: Transaction version.
    """

    # FIELDS

    type: ReceiptType
    version: ReceiptVersion
    network_type: OptionalNetworkType

    # OVERRIDABLE CLASSVARS
    # The following classvars should be re-implemented for each
    # base transaction type.

    # Registry of transaction types to models. Allows users to
    # register custom models to customize serialization/deserialization
    # logic.
    TYPE_MAP: typing.ClassVar[TypeMap]

    # Data to simplify the serialization/deserialization of
    # transactions of a given type. Provides high-level
    # routines, and also maps fields 1:1 to transaction DTO
    # keys.
    DTO: typing.ClassVar[DTOFormat]

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        """Validate receipt-specific fields in data-transfer object."""
        raise util.AbstractMethodError

    @classmethod
    def validate_dto_shared(cls, data: dict) -> bool:
        """Validate shared fields in data-transfer object."""
        raise util.AbstractMethodError

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        return (
            cls.validate_dto_shared(data)
            and cls.validate_dto_specific(data)
        )

    def to_dto_shared(
        self,
        network_type: OptionalNetworkType,
    ) -> dict:
        """Export shared receipt data to DTO. Internal use only."""
        raise util.AbstractMethodError

    def to_dto_specific(
        self,
        network_type: OptionalNetworkType,
    ) -> dict:
        """Export receipt-specific data to DTO. Internal use only."""
        raise util.AbstractMethodError

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        # Check network type matches transaction.
        if network_type is not None and network_type != self.network_type:
            raise ValueError('Network type does not match receipt.')

        # Save shared and specific receipt data.
        dto = self.to_dto_shared(self.network_type)
        dto.update(self.to_dto_specific(self.network_type))
        return dto

    def load_dto_shared(
        self,
        data: dict,
        network_type: OptionalNetworkType,
    ) -> None:
        """Load shared receipt data from DTO. Internal use only."""
        raise util.AbstractMethodError

    def load_dto_specific(
        self,
        data: dict,
        network_type: OptionalNetworkType,
    ) -> None:
        """Load receipt-specific data from DTO. Internal use only."""
        raise util.AbstractMethodError

    @classmethod
    def create_from_dto(
        cls: typing.Type[ReceiptBaseType],
        data: dict,
        network_type: OptionalNetworkType = None,
    ) -> ReceiptBaseType:
        """
        Deserialize object from DTO interchange format.

        If the cls is a direct subclass of `ReceiptBase`,
        determine the receipt type and therefore the correct class.
        If the cls is not a direct subclass of `TransactionBase`
        (and therefore a finalized receipt model), use the
        class directly.

        :param data: Receipt data in DTO interchange format.
        :param network_type: Network type.
        """

        # If we have a base class, find the correct derived class.
        if cls in ReceiptBase.__subclasses__():
            cls = cls.DTO.find_receipt(cls.TYPE_MAP, data)
        inst = typing.cast(ReceiptBaseType, cls.__new__(cls))

        # Validate after we find the correct subclass.
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        # Load and check the network type.
        # nt = cls.DTO.load_network_type(data)
        # if network_type is not None and network_type != nt:
        #    raise ValueError('Network type does not match receipt.')

        # Load shared and specific receipt data.
        inst.load_dto_shared(data, network_type)
        inst.load_dto_specific(data, network_type)
        return inst



"""
    rx
    ==

    Optional reactive extensions when RxPy is installed.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Import for side-effects, we have no other choice.
from . import from_async_iterable   # noqa: F401
from . import from_coroutine        # noqa: F401
from . import to_async_iterator     # noqa: F401
from . import to_future             # noqa: F401
from .async_generator import AsyncGenerator, AsyncGeneratorMixin
from .coroutine import Coroutine, CoroutineMixin

__all__ = [
    'AsyncGenerator',
    'AsyncGeneratorMixin',
    'Coroutine',
    'CoroutineMixin',
]



"""
    crypto
    ======

    Wrappers for PyCryptodome keccak hash functions.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Disable warnings for importing Keccak, PyCryptoDome is spuriously
# flagged for using legacy hashes, but we need to use those hashes
# for compatibility.
from __future__ import annotations
from Crypto.Hash import keccak      # nosec
import typing

# API


def keccak_224(data: typing.Optional[bytes] = None) -> keccak:
    """Returns a 224-bit keccak hash object; optionally initialized with a string."""

    return keccak.new(digest_bits=224, data=data)


def keccak_256(data: typing.Optional[bytes] = None) -> keccak:
    """Returns a 256-bit keccak hash object; optionally initialized with a string."""

    return keccak.new(digest_bits=256, data=data)


def keccak_384(data: typing.Optional[bytes] = None) -> keccak:
    """Returns a 384-bit keccak hash object; optionally initialized with a string."""

    return keccak.new(digest_bits=384, data=data)


def keccak_512(data: typing.Optional[bytes] = None) -> keccak:
    """Returns a 512-bit keccak hash object; optionally initialized with a string."""

    return keccak.new(digest_bits=512, data=data)



"""
    reactive
    ========

    Helpers to simplify interfacing with RxPy.

    If `RxPy` is installed, `@observable` makes `async def` functions
    return an `rx.Observable`. Otherwise, `@observable` does nothing.

    Example:
        .. code-block:: python

           >>> from xpxchain.util import observable
           >>> import asyncio
           >>> loop = asyncio.get_event_loop()
           >>> @observable
           ... async def f():
           ...     print("Called f()")
           ...     return 3
           >>> source = f().to_observable()
           >>> source.subscribe(lambda value: print("Received {0}".format(value)))
           >>> loop.run_until_complete(source)
           Called f()
           Received 3
           3

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import functools
import inspect

__all__ = ['observable']

try:            # noqa: C901
    # Have RxPy installed.
    from . import rx

    def observable_class(cls):
        """Decorate class so it supports a `to_observable` method."""

        has_aiter = hasattr(cls, '__aiter__') or hasattr(cls, '__anext__')
        has_await = hasattr(cls, '__await__')
        # Only our Observables should provide both, since they're
        # a blanket class. Disable this in release builds.
        assert not (has_aiter and has_await)

        # determine the proper mixin
        if has_aiter:
            mixin = rx.AsyncGeneratorMixin
        else:
            mixin = rx.CoroutineMixin

        class ToObservable(cls, mixin):
            pass

        ToObservable.__doc__ = cls.__doc__
        ToObservable.__module__ = cls.__module__
        ToObservable.__name__ = cls.__name__
        ToObservable.__qualname__ = cls.__qualname__

        return ToObservable

    def observable_routine(f):
        """Decorate async function so the result supports `to_observable`."""

        # Need to get the proper adapter.
        if inspect.isasyncgenfunction(f):
            adapter = rx.AsyncGenerator
        else:
            adapter = rx.Coroutine

        # Wrap methods and free-functions separately. If the first
        # argument is self, we want to get an event loop, if bound,
        # from the class.
        args = inspect.getfullargspec(f).args
        if args and args[0] == 'self':
            # Method
            @functools.wraps(f)
            def wrapper(self, *args, **kwds):
                loop = getattr(self, "loop", None)
                return adapter(f(self, *args, **kwds), loop=loop)
        else:
            # Function
            @functools.wraps(f)
            def wrapper(*args, **kwds):
                return adapter(f(*args, **kwds))

        return wrapper

    def observable(value):
        """Decorates a function or class to allow `to_observable`."""

        if inspect.isclass(value):
            return observable_class(value)
        elif inspect.isroutine(value):
            return observable_routine(value)
        raise TypeError("Observable must be created from class or routine.")

except ImportError:
    # RxPy is not installed.

    def observable(value):
        """Dummy decorator that returns whatever was passed to it."""
        return value



import binascii
import string

from xpxchain import models
from tests import harness


class TestMosaic(harness.TestCase):

    @harness.randomize
    def test_valid(self, id: harness.U64, amount: harness.U64):
        network_type = models.NetworkType.MIJIN_TEST
        model = models.Mosaic(models.MosaicId(id), amount)
        dto = model.to_dto(network_type)
        self.assertEqual(model, models.Mosaic.create_from_dto(dto, network_type))

    @harness.randomize(
        id={'min_value': -1 << 32, 'max_value': -1},
        amount={'min_value': -1 << 32, 'max_value': -1},
    )
    def test_invalid(self, id: int, amount: int):
        network_type = models.NetworkType.MIJIN_TEST
        model = models.Mosaic(models.MosaicId(id), amount)
        with self.assertRaises(ArithmeticError):
            model.to_dto(network_type)


class TestMosaicId(harness.TestCase):

    @harness.randomize
    def test_valid(self, id: harness.U64):
        network_type = models.NetworkType.MIJIN_TEST
        model = models.Mosaic(models.MosaicId(id), 1)
        dto = model.to_dto(network_type)
        self.assertEqual(model, models.Mosaic.create_from_dto(dto, network_type))

    @harness.randomize(id={'min_value': -1 << 32, 'max_value': -1})
    def test_invalid(self, id: int):
        network_type = models.NetworkType.MIJIN_TEST
        model = models.Mosaic(models.MosaicId(id), 1)
        with self.assertRaises(ArithmeticError):
            model.to_dto(network_type)


class TestMosaicNonce(harness.TestCase):

    @harness.randomize(nonce={'fixed_length': 4})
    def test_valid_bytes(self, nonce: bytes):
        network_type = models.NetworkType.MIJIN_TEST
        model = models.MosaicNonce(nonce)
        dto = model.to_dto(network_type)
        self.assertEqual(model, models.MosaicNonce.create_from_dto(dto, network_type))
        catbuffer = model.to_catbuffer(network_type)
        self.assertEqual(model, models.MosaicNonce.create_from_catbuffer(catbuffer, network_type))

    @harness.randomize(nonce={'letters': string.hexdigits, 'fixed_length': 8})
    def test_valid_str(self, nonce: str):
        network_type = models.NetworkType.MIJIN_TEST
        model = models.MosaicNonce(nonce)
        dto = model.to_dto(network_type)
        self.assertEqual(model, models.MosaicNonce.create_from_dto(dto, network_type))
        catbuffer = model.to_catbuffer(network_type)
        self.assertEqual(model, models.MosaicNonce.create_from_catbuffer(catbuffer, network_type))

    @harness.randomize
    def test_valid_int(self, nonce: harness.U32):
        network_type = models.NetworkType.MIJIN_TEST
        model = models.MosaicNonce(nonce)
        dto = model.to_dto(network_type)
        self.assertEqual(model, models.MosaicNonce.create_from_dto(dto, network_type))
        catbuffer = model.to_catbuffer(network_type)
        self.assertEqual(model, models.MosaicNonce.create_from_catbuffer(catbuffer, network_type))

    @harness.randomize(nonce={'min_length': 0, 'max_length': 3})
    def test_invalid_bytes(self, nonce: bytes):
        with self.assertRaises(ValueError):
            models.MosaicNonce(nonce)

    @harness.randomize(nonce={'letters': string.hexdigits, 'min_length': 0, 'max_length': 7})
    def test_invalid_str(self, nonce: str):
        with self.assertRaises((ValueError, binascii.Error)):
            models.MosaicNonce(nonce)

    @harness.randomize(nonce={'min_value': -1 << 32, 'max_value': -1})
    def test_invalid_int(self, nonce: int):
        with self.assertRaises(OverflowError):
            models.MosaicNonce(nonce)


class TestMosaicProperties(harness.TestCase):

    @harness.randomize(divisibility={'min_value': 0, 'max_value': 6})
    def test_create(
        self,
        supply_mutable: bool,
        transferable: bool,
        levy_mutable: bool,
        divisibility: int,
        duration: harness.U64,
    ):
        properties = models.MosaicProperties.create(
            supply_mutable=supply_mutable,
            transferable=transferable,
            levy_mutable=levy_mutable,
            divisibility=divisibility,
            duration=duration,
        )
        self.assertEqual(properties.supply_mutable, supply_mutable)
        self.assertEqual(properties.transferable, transferable)
        self.assertEqual(properties.levy_mutable, levy_mutable)
        self.assertLessEqual(properties.flags, 7)
        self.assertEqual(properties.divisibility, divisibility)
        self.assertEqual(properties.duration, duration)
        self.assertEqual(properties, models.MosaicProperties.create_from_dto(properties.to_dto()))

    @harness.randomize(divisibility={'min_value': 7, 'max_value': 1 << 31})
    def test_invalid_divisibility(self, divisibility: int):
        with self.assertRaises(ValueError):
            models.MosaicProperties.create(divisibility=divisibility)
        with self.assertRaises(ValueError):
            models.MosaicProperties.create(divisibility=-divisibility)


# TODO(ahuszagh) Add supply type...
# TODO(ahuszagh) Add network currency type...
# TODO(ahuszagh) Add network harvest type...



"""
    account_property_modification
    =============================

    Account property modification type and value.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from ..account.account_property import PropertyValue
from ..account.address import Address
from ..account.property_modification_type import PropertyModificationType
from ..blockchain.network_type import OptionalNetworkType
from ..mosaic.mosaic_id import MosaicId
from ..transaction.transaction_type import TransactionType
from ... import util

__all__ = ['AccountPropertyModification']

DTOType = typing.Union[
    str,
    typing.Sequence[int],
    int
]


def to_dto(value: PropertyValue) -> DTOType:
    """Serialize property modification value to DTO."""

    if isinstance(value, Address):
        return value.address
    elif isinstance(value, MosaicId):
        return util.u64_to_dto(int(value.id))
    elif isinstance(value, TransactionType):
        return value.to_dto()
    else:
        raise NotImplementedError


def from_dto(data: DTOType) -> PropertyValue:
    """Load property modification value from DTO."""

    if (isinstance(data, str) and (len(data) == 50)):
        return Address.create_from_encoded(data)
    elif (isinstance(data, str) and (len(data) == 40)):
        return Address.create_from_raw_address(data)
    elif isinstance(data, list):
        return MosaicId(util.u64_from_dto(data))
    elif isinstance(data, int):
        return TransactionType.create_from_dto(data)
    else:
        raise NotImplementedError


@util.inherit_doc
@util.dataclass(frozen=True)
class AccountPropertyModification(util.DTO):
    """
    Account property modification type and value.

    :param modification_type: Property modification type.
    :param value: Modification value.
    """

    modification_type: PropertyModificationType
    value: PropertyValue

    def is_address(self) -> bool:
        """Determine if the modification value type is an address."""
        return isinstance(self.value, Address)

    def is_mosaic(self) -> bool:
        """Determine if the modification value type is a mosaic ID."""
        return isinstance(self.value, MosaicId)

    def is_entity_type(self) -> bool:
        """Determine if the modification value type is an entity type."""
        return isinstance(self.value, TransactionType)

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys = {'type', 'value'}
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, required_keys)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'type': self.modification_type.to_dto(network_type),
            'value': to_dto(self.value),
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        type = data['type']
        return cls(
            modification_type=PropertyModificationType.create_from_dto(type),
            value=from_dto(data['value']),
        )



"""
    async_generator
    ===============

    Wrapper for asynchronous generators.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing
import rx


class AsyncGenerator:
    """Asynchronous generator wrapper type."""

    def __init__(self, value, loop=None):
        self._value = value
        self._loop = loop
        self._observable = None
        # In case we get an asynchronous iterable, which does not
        # support __anext__. If we get an asynchronous iterator,
        # or an asychronous generator iterator, use.
        if not hasattr(self._value, '__anext__'):
            self._value = self._value.__aiter__()

    async def aclose(self) -> None:
        """Raise GeneratorExit inside generator."""
        await self._value.aclose()

    async def asend(self, arg):
        """Send `arg` into generator."""
        return await self._value.asend(arg)

    async def throw(self, exception_type, *args):
        """Raise exception in generator."""
        return await self._value.athrow(exception_type, *args)

    def to_observable(self):
        """Export awaitable object to Observable."""
        if self._observable is None:
            self._observable = rx.Observable.from_async_iterable(
                self._value,
                loop=self._loop
            )
        return self._observable

    @property
    def ag_await(self):
        return self._value.ag_await

    @property
    def ag_code(self):
        return self._value.ag_code

    @property
    def ag_frame(self):
        return self._value.ag_frame

    @property
    def ag_running(self):
        return self._value.ag_running

    def __aiter__(self):
        return self

    def __anext__(self):
        if self._observable is None:
            return self._value.__anext__()
        return self._observable.__anext__()


class AsyncGeneratorMixin(typing.AsyncIterable):
    """Mixin for async iterable classes to support `to_observable`."""

    def to_observable(self):
        """Export async iterable object to Observable."""
        loop = getattr(self, 'loop', None)
        return rx.Observable.from_async_iterable(
            self.__aiter__(),
            loop=loop
        )
